#include <system.h>

#pragma DATA _CONFIG, _INTRC_OSC_NOCLKOUT & _CP_OFF & _CPD_OFF & _PWRTE_ON & _WDT_ON & _MCLRE_OFF & _BODEN_ON
#pragma CLOCK_FREQ 4000000

// I/O pin assignments
#define SERVO		GP5
#define MAS_SIGNAL	GP1
#define MC2_SIGNAL	GP2
#define LED 		GP4
#define SELECT_PIC	GP0
#define BUTTON		GP3

// below positions are defined in 100ths of ms (also known as 10us increments)
#define WING_CLOSED_POS	217 /* max tightness: 1.104ms, somwhat tight: ~1.150 */
#define WING_OPEN_POS	158 /* 1.582ms, empirically */

// these are arbitrary times.. 20000 is approximately half a second at the time of this writing
#define BUTTON_TIMEOUT 20000
#define SIGNAL_TIMEOUT 2000
//#define LED_BLINK_RATE 5000


#define STATE_CLOSED	0
#define STATE_CLOSING	1
#define STATE_OPEN		2

#define NUM_CLOSING_STEPS	1

unsigned char state = STATE_CLOSED; // current state of the system (for button)
unsigned char closingStep = 0; // which state in the closing sequence we are at
unsigned short LEDOnTime = 1000; // blink rate for the LED (in main loop cycles)
unsigned short LEDOffTime = 20000;

unsigned char desiredServoPos = WING_CLOSED_POS; // the desired position of the servo
unsigned char servoPos = WING_CLOSED_POS;

unsigned char pulseCounter = 0;
bool servoStepped = false;

/*
	Outputs a pulse on ServoPin approx every 30ms to control a servo.  Pulse width should be between 1 and 2ms,
	which corresponds to a desiredServoPos value of 100 to 200 (10us increments).
*/

void main()
{
	unsigned short counter = 0;
	unsigned short buttonTimeout = 0;
	unsigned short signalTimeout = 0;
	trisio = 0; // all pins outputs
	ansel = 0; // disable A/D
	gpio = 0; // all pins low
	
	cmcon |= 0x7; // turn off comparator
	
	// retrieve ocscal and set it
	asm
	{
		bsf _status, RP0  	; Bank 1
		call 0x3ff 	; Get the cal value
		movwf _osccal 	; Calibrate
		bcf _status, RP0 	; Bank 0
	}
	
	// set inputs
	set_bit(trisio, BUTTON);
	set_bit(trisio, MC2_SIGNAL);
	set_bit(trisio, MAS_SIGNAL);
	
	clear_bit(option_reg, NOT_GPPU); // enable weak pull up resistors
	
	// set up timer
	clear_bit(option_reg, T0CS); // TMR0 operates on internal clock
	clear_wdt();
	clear_bit(option_reg, PSA); // prescaler is assigned to TMR0
	
	// set prescaler to 110 (1:128, approx every 30ms)
	set_bit(option_reg, PS2);
	set_bit(option_reg, PS1);
	clear_bit(option_reg, PS0);
	
	set_bit(intcon, GIE);
	
	set_bit(intcon, T0IE); // enable TMR0 
	
	set_bit(trisio, GPIO2); // make IO pin 2 an input
	
	set_bit(gpio, SELECT_PIC); // make the PIC the output
	
	// just for safety (in case of reset?)
	desiredServoPos = WING_CLOSED_POS;
	state = STATE_CLOSED;
	
	while (1)
	{
		// check for signal from either MAS or MC2
		if (test_bit(gpio, MAS_SIGNAL) || !test_bit(gpio, MC2_SIGNAL))
			++signalTimeout;
		else
			signalTimeout = 0;
			
		// check to see if button is pressed
		if (gpio.BUTTON == 0)
			++buttonTimeout;
		else
			buttonTimeout = 0;
			
		// if button has been pressed long enough to trigger, toggle wing position
		if (buttonTimeout >= BUTTON_TIMEOUT)
		{
			switch (state) // state machine to figure out how to set things.  since the button is the only input capable of getting into these states, the state machine is in the button handler
			{
			case STATE_CLOSED: // move to open
				desiredServoPos = WING_OPEN_POS;
				state = STATE_OPEN;
				LEDOnTime = 1200;
				LEDOffTime = 1200;
				break;
				
			case STATE_OPEN: // move to closing
				state = STATE_CLOSING;
				LEDOnTime = 6000;
				LEDOffTime = 6000;
				closingStep = 0;
				// fall through intended			
				
			case STATE_CLOSING: // move to next closed state or closed
				desiredServoPos = (((WING_CLOSED_POS - WING_OPEN_POS) * (closingStep)) / NUM_CLOSING_STEPS) + WING_OPEN_POS;
				++closingStep;
			
				if (closingStep >= NUM_CLOSING_STEPS) // move to closed state
				{
					state = STATE_CLOSED;
					desiredServoPos = WING_CLOSED_POS;
					LEDOnTime = 1000;
					LEDOffTime = 20000;
				}
				break;
				
			default: // default to open as a saftey feature
				state = STATE_OPEN;
				desiredServoPos = WING_OPEN_POS;
				break;
			}
			
			buttonTimeout = 0;
		}
		else if (signalTimeout >= SIGNAL_TIMEOUT) // if signal has been active long enough to trigger, open wings
		{
			desiredServoPos = WING_OPEN_POS;
			state = STATE_OPEN;
		}
			
		// blink the LED
		++counter;
		if (gpio.LED == 1 && counter >= LEDOnTime)
		{
			gpio.LED = 0;
			counter = 0;
		}
		else if (gpio.LED == 0 && counter >= LEDOffTime)
		{
			gpio.LED = 1;
			counter = 0;
		}
		
		clear_wdt();
		
		if ((pulseCounter & 0x1) == 0 && !servoStepped)
		{
			servoStepped = true;
			if (servoPos < desiredServoPos)
			{
				servoPos += 2;
				
				if (servoPos > desiredServoPos)
					servoPos = desiredServoPos;
			}
			else if (servoPos > desiredServoPos)
			{
				servoPos -= 2;
				
				if (servoPos < desiredServoPos)
					servoPos = desiredServoPos;
			}
		}
		else if (pulseCounter & 0x1)
			servoStepped = false;
	}
	
}

void interrupt( void )
{
	if (test_bit(intcon, T0IF))
	{
		tmr0 = 90;
		++pulseCounter;
		clear_bit(intcon, T0IF);  //clear TMR0 overflow flag
		set_bit(gpio, SERVO);
		delay_10us(servoPos);
		clear_bit(gpio, SERVO);
	}
}
